# iOS Integration

>Note: the following documents assume that you have some experience in iOS development.

This tutorial shows how to integrate Hippy into an iOS project.

# Integrate with Pod

1. install [cocoapods](https://cocoapods.org/), Hippy iOS SDK [version query](https://cocoapods.org/pods/hippy)

2. Create a podfile file under user-defined project directory with the following text

    ```text
    # keep podfile directory structure
    install! "cocoapods", :preserve_pod_file_structure => true
    platform :ios, '8.0'
    # replace TargetName with user's project name
    target TargetName do
        # use latest version of Hippy
        pod 'hippy'
        # if you want to assign Hippy version, say, 2.0.0
        #pod 'hippy', '2.0.0'
    end
    ```

3. Configure force load options

    Hippy uses reflection a lot. If you compile Hippy code as a static link Library, the code that is not explicitly called will be stripped by the compiler.
    Therefore, you must set `force load` to force loading all symbols of the Hippy static library.
   
    >  force load is deleted since ver 2.13.0。If static link Library access is used, app needs to configure it by itself.

    There are multiple ways to simulate `force load` , choose one below and adapt accordingly.

    * Configure `*-force_load "${PODS_CONFIGURATION_BUILD_DIR}/hippy/libhippy.a"*` in the `Other Linker Flags` of the build settings of the target corresponding to the main project.

    * Add `post_install hook` in the project's podfile configuration file, and manually add `force load` to the xcconfig

    * fork Hippy source code, change `hippy.podspec`, add the config below to `user_target`, and include this source code.

    ```text
    s.user_target_xcconfig = {'OTHER_LDFLAGS' => '-force_load "${PODS_CONFIGURATION_BUILD_DIR}/hippy/libhippy.a"'}
    ```

4. Execute in the command line

    ```text
    pod install
    ```

5. open the project by the `.xcworkspace` file generated by cocoapods.

# Direct Integration from Source Code

1. Download the Hippy iOS SDK source code from GitHub and drag the ios/sdk folder and the core folder into the project

2. removes references to `core/js` folder.

   >The core/js folder contains js files that do not participate in compilation

3. Remove references to `core/napi/v8` folder.

   >The core folder involves JS references for different platforms (iOS/Android), and iOS uses JSCore

4. set the *User Header Search Paths* as the path of the core folder in `xcode build settings`

 >Assume the core folder path is`~/documents/project/hippy/demo/core`, you should set `~/documents/project/hippy/demo/` instead of`~/documents/project/hippy/demo/core`

# Write Code to Start Debugging or Load Business Code

Hippy provides a code-splitting loading interface and a non-code-splitting loading interface. All service packages are carried through `HippyRootView`. Creating a business is to create a `RootView`.

## Using the code-splitting load interface

``` objectivec
/** This method is suitable for the following scenarios.
 * Prepare the JS environment and load package 1 before the business is started, and load package 2 when the business is started to reduce package loading time
 * We suggest package 1 as the base package, which is not related to the business and contains only some common base components, common to all businesses
 * Package 2 is loaded as business code
*/

// Load package 1 address first to create execution environment
//commonBundlePath being package 1's path
NSURL * commonBundlePath = getCommonBundlePath();
HippyBridge *bridge = [[HippyBridge alloc] initWithBundleURL: commonBundlePath
                                                moduleProvider: nil
                                                launchOptions: nil];

// Create rootview by bridge and package2 address
- (instancetype)initWithBridge:(HippyBridge *)bridge  
    businessURL:(NSURL *)businessURL // business package address
    moduleName:(NSString *)moduleName // business package startup function name
    initialProperties:(NSDictionary *)initialProperties // initialize parameters
    shareOptions:(NSDictionary *)shareOptions // configuration parameters (advanced)
    isDebugMode:(BOOL)isDebugMode // whether debug mode
    delegate:(id<HippyRootViewDelegate>)delegate // rootview load callback

```

## Using the non-code-splitting load interface

``` objectivec
Subpackage loading interface - (instancetype)initWithBundleURL:(NSURL *)bundleURL // package address
    moduleName:(NSString *)moduleName // business package startup function name
    initialProperties:(NSDictionary *)initialProperties // initialize parameters
    shareOptions:(NSDictionary *)shareOptions // configuration parameters (advanced)
    isDebugMode:(BOOL)isDebugMode // whether debug mode
    delegate:(id <HippyRootViewDelegate>)delegate // rootview load callback
```

!> Regardless of whether rootview is initialized with or without code-splitting, if **isdebugmode** is YES, all parameters will be ignored and the test dataset is loaded directly using the npm local service. Using code-splitting loading can be combined with a series of strategies, such as pre-loading bridge, global single bridge, and so on, to optimize page opening speed.
